/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/* NetworkManager -- Network link manager
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Copyright 2011 - 2012 Red Hat, Inc.
 */

#include "config.h"

#include <glib.h>
#include <glib/gi18n.h>

#include <netinet/ether.h>
#include <errno.h>
#include <stdlib.h>

#include "nm-device-bond.h"
#include "nm-device-interface.h"
#include "nm-logging.h"
#include "nm-properties-changed-signal.h"
#include "nm-utils.h"
#include "NetworkManagerUtils.h"
#include "nm-device-private.h"
#include "nm-netlink-monitor.h"
#include "nm-dbus-glib-types.h"
#include "nm-system.h"
#include "nm-dhcp-manager.h"

#include "nm-device-bond-glue.h"

G_DEFINE_TYPE (NMDeviceBond, nm_device_bond, NM_TYPE_DEVICE_WIRED)

#define NM_DEVICE_BOND_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NM_TYPE_DEVICE_BOND, NMDeviceBondPrivate))

#define NM_BOND_ERROR (nm_bond_error_quark ())

typedef struct {
	gboolean ip4_waiting;
	gboolean ip6_waiting;
	gboolean carrier_waiting;

	guint ip_retry_id;
} NMDeviceBondPrivate;

enum {
	PROPERTIES_CHANGED,

	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

enum {
	PROP_0,
	PROP_HW_ADDRESS,
	PROP_CARRIER,
	PROP_SLAVES,

	LAST_PROP
};

/******************************************************************/

typedef enum {
	NM_BOND_ERROR_CONNECTION_NOT_BOND = 0,
	NM_BOND_ERROR_CONNECTION_INVALID,
	NM_BOND_ERROR_CONNECTION_INCOMPATIBLE,
} NMBondError;

#define NM_BOND_ERROR (nm_bond_error_quark ())
#define NM_TYPE_BOND_ERROR (nm_bond_error_get_type ())

static GQuark
nm_bond_error_quark (void)
{
	static GQuark quark = 0;
	if (!quark)
		quark = g_quark_from_static_string ("nm-bond-error");
	return quark;
}

/* This should really be standard. */
#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }

static GType
nm_bond_error_get_type (void)
{
	static GType etype = 0;

	if (etype == 0) {
		static const GEnumValue values[] = {
			/* Connection was not a bond connection. */
			ENUM_ENTRY (NM_BOND_ERROR_CONNECTION_NOT_BOND, "ConnectionNotBond"),
			/* Connection was not a valid BT connection. */
			ENUM_ENTRY (NM_BOND_ERROR_CONNECTION_INVALID, "ConnectionInvalid"),
			/* Connection does not apply to this device. */
			ENUM_ENTRY (NM_BOND_ERROR_CONNECTION_INCOMPATIBLE, "ConnectionIncompatible"),
			{ 0, 0, 0 }
		};
		etype = g_enum_register_static ("NMBondError", values);
	}
	return etype;
}

/******************************************************************/

static void
device_state_changed (NMDevice *device,
                      NMDeviceState new_state,
                      NMDeviceState old_state,
                      NMDeviceStateReason reason,
                      gpointer user_data)
{
	NMDeviceBondPrivate *priv = NM_DEVICE_BOND_GET_PRIVATE (device);
	const char *iface = nm_device_get_ip_iface (device);

	if (new_state != NM_DEVICE_STATE_IP_CONFIG) {
		if (priv->ip_retry_id) {
			g_source_remove (priv->ip_retry_id);
			priv->ip_retry_id = 0;
		}
	}

	/* Remove all bond slaves to ensure we start with a clear state */
	if (new_state == NM_DEVICE_STATE_UNAVAILABLE) {
		NMDHCPManager *dhcp_manager;
		char *contents, *path;

		path = g_strdup_printf ("/sys/class/net/%s/bonding/slaves", iface);
		g_assert (path);
		if (g_file_get_contents (path, &contents, 0, NULL)) {
			char **items, **iter;
			int slave_ifindex;

			items = g_strsplit_set (contents, "\n\r", -1);
			for (iter = items; iter && *iter; iter++) {
				if (*iter[0]) {
					slave_ifindex = nm_netlink_iface_to_index (*iter);
					if (slave_ifindex > 0) {
						nm_log_info (LOGD_DEVICE, "(%s): detaching slave %s",
							        iface, *iter);
						nm_system_bond_release (nm_device_get_ip_ifindex (device),
							                    iface,
							                    slave_ifindex,
							                    *iter);
					}
				}
			}
			if (items)
				g_strfreev (items);
			g_free (contents);
		}
		g_free (path);

		/* Stop any existing dhcp client */
		dhcp_manager = nm_dhcp_manager_get ();
		nm_dhcp_manager_stop_existing (dhcp_manager, iface, FALSE);
		g_object_unref (dhcp_manager);
	}
}

static void
carrier_action (NMDeviceWired *self, NMDeviceState state, gboolean carrier)
{
	/* Carrier can't be used to signal availability of the bond master because
	 * the bond's carrier follows the slaves' carriers.  So carrier gets
	 * ignored when determining whether or not the device can be activated.
	 *
	 * Second, just because all slaves have been removed or have lost carrier
	 * does not mean the master should be deactivated.  This could be due to
	 * user addition/removal of slaves, and is also normal operation with some
	 * failover modes.
	 *
	 * For these reasons, carrier changes are effectively ignored by overriding
	 * the parent class' carrier handling and doing nothing.
	 */
}

static void
real_update_hw_address (NMDevice *dev)
{
	const guint8 *hw_addr;
	guint8 old_addr[NM_UTILS_HWADDR_LEN_MAX];
	int addrtype, addrlen;

	addrtype = nm_device_wired_get_hwaddr_type (NM_DEVICE_WIRED (dev));
	g_assert (addrtype >= 0);
	addrlen = nm_utils_hwaddr_len (addrtype);
	g_assert (addrlen > 0);

	hw_addr = nm_device_wired_get_hwaddr (NM_DEVICE_WIRED (dev));
	memcpy (old_addr, hw_addr, addrlen);

	NM_DEVICE_CLASS (nm_device_bond_parent_class)->update_hw_address (dev);

	hw_addr = nm_device_wired_get_hwaddr (NM_DEVICE_WIRED (dev));
	if (memcmp (old_addr, hw_addr, addrlen))
		g_object_notify (G_OBJECT (dev), NM_DEVICE_BOND_HW_ADDRESS);
}

static guint32
real_get_generic_capabilities (NMDevice *dev)
{
	return NM_DEVICE_CAP_CARRIER_DETECT | NM_DEVICE_CAP_NM_SUPPORTED;
}

static gboolean
is_available (NMDevice *dev)
{
	/* Carrier doesn't really matter for bond interfaces, we only care if it's IFF_UP */
	if (NM_DEVICE_GET_CLASS (dev)->hw_is_up)
		return NM_DEVICE_GET_CLASS (dev)->hw_is_up (dev);
	return FALSE;
}

static gboolean
match_bond_connection (NMDevice *device, NMConnection *connection, GError **error)
{
	const char *iface;
	NMSettingBond *s_bond;

	s_bond = nm_connection_get_setting_bond (connection);
	if (!s_bond || !nm_connection_is_type (connection, NM_SETTING_BOND_SETTING_NAME)) {
		g_set_error (error, NM_BOND_ERROR, NM_BOND_ERROR_CONNECTION_NOT_BOND,
		             "The connection was not a bond connection.");
		return FALSE;
	}

	/* Bond connections must specify the virtual interface name */
	iface = nm_connection_get_virtual_iface_name (connection);
	if (!iface || strcmp (nm_device_get_iface (device), iface)) {
		g_set_error (error, NM_BOND_ERROR, NM_BOND_ERROR_CONNECTION_NOT_BOND,
		             "The bond connection virtual interface name did not match.");
		return FALSE;
	}

	/* FIXME: match bond properties like mode, etc? */

	return TRUE;
}

static NMConnection *
real_get_best_auto_connection (NMDevice *dev,
                               GSList *connections,
                               char **specific_object)
{
	GSList *iter;

	for (iter = connections; iter; iter = g_slist_next (iter)) {
		NMConnection *connection = NM_CONNECTION (iter->data);
		NMSettingConnection *s_con;

		s_con = nm_connection_get_setting_connection (connection);
		g_assert (s_con);
		if (   nm_setting_connection_get_autoconnect (s_con)
		    && match_bond_connection (dev, connection, NULL))
			return connection;
	}
	return NULL;
}

static gboolean
real_check_connection_compatible (NMDevice *device,
                                  NMConnection *connection,
                                  GError **error)
{
	return match_bond_connection (device, connection, error);
}

static gboolean
spec_match_list (NMDevice *device, const GSList *specs)
{
	char *hwaddr;
	gboolean matched;

	hwaddr = nm_utils_hwaddr_ntoa (nm_device_wired_get_hwaddr (NM_DEVICE_WIRED (device)), ARPHRD_ETHER);
	matched = nm_match_spec_hwaddr (specs, hwaddr);
	g_free (hwaddr);

	if (!matched)
		matched = nm_match_spec_interface_name (specs, nm_device_get_iface (device));

	return matched;
}

static gboolean
bond_match_config (NMDevice *self, NMConnection *connection)
{
	NMSettingBond *s_bond;
	const char *ifname;

	s_bond = nm_connection_get_setting_bond (connection);
	if (!s_bond)
		return FALSE;

	/* Interface name */
	ifname = nm_setting_bond_get_interface_name (s_bond);
	if (g_strcmp0 (ifname, nm_device_get_ip_iface (self)) != 0)
		return FALSE;

	return TRUE;
}

static NMConnection *
connection_match_config (NMDevice *self, const GSList *connections)
{
	const GSList *iter;
	GSList *bond_matches;
	NMConnection *match;

	/* First narrow @connections down to those that match in their
	 * NMSettingBond configuration.
	 */
	bond_matches = NULL;
	for (iter = connections; iter; iter = iter->next) {
		NMConnection *candidate = NM_CONNECTION (iter->data);

		if (!nm_connection_is_type (candidate, NM_SETTING_BOND_SETTING_NAME))
			continue;
		if (!bond_match_config (self, candidate))
			continue;

		bond_matches = g_slist_prepend (bond_matches, candidate);
	}

	/* Now pass those to the super method, which will check IP config */
	bond_matches = g_slist_reverse (bond_matches);
	match = NM_DEVICE_CLASS (nm_device_bond_parent_class)->connection_match_config (self, bond_matches);
	g_slist_free (bond_matches);

	return match;
}

/******************************************************************/

static gboolean
set_bond_attr (const char *iface, const char *attr, const char *value)
{
	char file[FILENAME_MAX];
	gboolean ret;

	snprintf (file, sizeof (file), "/sys/class/net/%s/bonding/%s", iface, attr);
	ret = nm_utils_do_sysctl (file, value);
	if (!ret) {
		nm_log_warn (LOGD_HW, "(%s): failed to set bonding attribute "
		             "'%s' to '%s': %d", iface, attr, value, errno);
	}
	return ret;
}

static void
set_arp_targets (const char *iface,
                 const char *value,
                 const char *delim,
                 const char *prefix)
{
	char **items, **iter, *tmp;

	items = g_strsplit_set (value, delim, 0);
	for (iter = items; iter && *iter; iter++) {
		if (*iter[0]) {
			tmp = g_strdup_printf ("%s%s", prefix, *iter);
			set_bond_attr (iface, "arp_ip_target", tmp);
			g_free (tmp);
		}
	}
	g_strfreev (items);
}

static void
set_simple_option (const char *iface,
                   const char *attr,
                   NMSettingBond *s_bond,
                   const char *opt)
{
	const char *value, *def;

	value = nm_setting_bond_get_option_by_name (s_bond, opt);
	def = nm_setting_bond_get_option_default (s_bond, opt);
	set_bond_attr (iface, attr, value ? value : def);
}

static NMActStageReturn
apply_bonding_config (NMDevice *device)
{
	NMConnection *connection;
	NMSettingBond *s_bond;
	const char *iface = nm_device_get_ip_iface (device);
	const char *mode, *value;
	char *path, *contents;
	gboolean set_arp_interval = TRUE;

	/* Option restrictions:
	 *
	 * arp_interval conflicts miimon > 0
	 * arp_interval conflicts [ alb, tlb ]
	 * arp_validate needs [ active-backup ]
	 * downdelay needs miimon
	 * updelay needs miimon
	 * primary needs [ active-backup, tlb, alb ]
	 *
	 * clearing miimon requires that arp_interval be 0, but clearing
	 *     arp_interval doesn't require miimon to be 0
	 */

	connection = nm_device_get_connection (device);
	g_assert (connection);
	s_bond = nm_connection_get_setting_bond (connection);
	g_assert (s_bond);

	mode = nm_setting_bond_get_option_by_name (s_bond, NM_SETTING_BOND_OPTION_MODE);
	if (mode == NULL)
		mode = "balance-rr";

	value = nm_setting_bond_get_option_by_name (s_bond, NM_SETTING_BOND_OPTION_MIIMON);
	if (value && atoi (value)) {
		/* clear arp interval */
		set_bond_attr (iface, "arp_interval", "0");
		set_arp_interval = FALSE;

		set_bond_attr (iface, "miimon", value);
		set_simple_option (iface, "updelay", s_bond, NM_SETTING_BOND_OPTION_UPDELAY);
		set_simple_option (iface, "downdelay", s_bond, NM_SETTING_BOND_OPTION_DOWNDELAY);
	} else if (!value) {
		/* If not given, and arp_interval is not given, default to 100 */
		long int val_int;
		char *end;

		value = nm_setting_bond_get_option_by_name (s_bond, NM_SETTING_BOND_OPTION_ARP_INTERVAL);
		errno = 0;
		val_int = strtol (value ? value : "0", &end, 10);
		if (!value || (val_int == 0 && errno == 0 && *end == '\0'))
			set_bond_attr (iface, "miimon", "100");
	}

	/* The stuff after 'mode' requires the given mode or doesn't care */
	set_bond_attr (iface, "mode", mode);

	/* arp_interval not compatible with ALB, TLB */
	if (g_strcmp0 (mode, "balance-alb") == 0 || g_strcmp0 (mode, "balance-tlb") == 0)
		set_arp_interval = FALSE;

	if (set_arp_interval) {
		set_simple_option (iface, "arp_interval", s_bond, NM_SETTING_BOND_OPTION_ARP_INTERVAL);

		/* Just let miimon get cleared automatically; even setting miimon to
		 * 0 (disabled) clears arp_interval.
		 */
	}

	value = nm_setting_bond_get_option_by_name (s_bond, NM_SETTING_BOND_OPTION_ARP_VALIDATE);
	/* arp_validate > 0 only valid in active-backup mode */
	if (   value
	    && g_strcmp0 (value, "0") != 0
	    && g_strcmp0 (value, "none") != 0
	    && g_strcmp0 (mode, "active-backup") == 0)
		set_bond_attr (iface, "arp_validate", value);
	else
		set_bond_attr (iface, "arp_validate", "0");

	if (   g_strcmp0 (mode, "active-backup") == 0
	    || g_strcmp0 (mode, "balance-alb") == 0
	    || g_strcmp0 (mode, "balance-tlb") == 0) {
		value = nm_setting_bond_get_option_by_name (s_bond, NM_SETTING_BOND_OPTION_PRIMARY);
		set_bond_attr (iface, "primary", value ? value : "");
	}

	/* Clear ARP targets */
	path = g_strdup_printf ("/sys/class/net/%s/bonding/arp_ip_target", iface);
	if (g_file_get_contents (path, &contents, NULL, NULL)) {
		set_arp_targets (iface, contents, " \n", "-");
		g_free (contents);
	}
	g_free (path);

	/* Add new ARP targets */
	value = nm_setting_bond_get_option_by_name (s_bond, NM_SETTING_BOND_OPTION_ARP_IP_TARGET);
	if (value)
		set_arp_targets (iface, value, ",", "+");

	set_simple_option (iface, "primary_reselect", s_bond, NM_SETTING_BOND_OPTION_PRIMARY_RESELECT);
	set_simple_option (iface, "fail_over_mac", s_bond, NM_SETTING_BOND_OPTION_FAIL_OVER_MAC);
	set_simple_option (iface, "use_carrier", s_bond, NM_SETTING_BOND_OPTION_USE_CARRIER);
	set_simple_option (iface, "ad_select", s_bond, NM_SETTING_BOND_OPTION_AD_SELECT);
	set_simple_option (iface, "xmit_hash_policy", s_bond, NM_SETTING_BOND_OPTION_XMIT_HASH_POLICY);
	set_simple_option (iface, "resend_igmp", s_bond, NM_SETTING_BOND_OPTION_RESEND_IGMP);

	return NM_ACT_STAGE_RETURN_SUCCESS;
}


static NMActStageReturn
real_act_stage1_prepare (NMDevice *dev, NMDeviceStateReason *reason)
{
	NMDeviceBondPrivate *priv = NM_DEVICE_BOND_GET_PRIVATE (dev);
	NMActStageReturn ret = NM_ACT_STAGE_RETURN_SUCCESS;
	gboolean no_firmware = FALSE;

	g_return_val_if_fail (reason != NULL, NM_ACT_STAGE_RETURN_FAILURE);

	priv->ip4_waiting = FALSE;
	priv->ip6_waiting = FALSE;
	priv->carrier_waiting = FALSE;

	ret = NM_DEVICE_CLASS (nm_device_bond_parent_class)->act_stage1_prepare (dev, reason);
	if (ret != NM_ACT_STAGE_RETURN_SUCCESS)
		return ret;

	/* Interface must be down to set bond options */
	nm_device_hw_take_down (dev, TRUE);
	ret = apply_bonding_config (dev);
	nm_device_hw_bring_up (dev, TRUE, &no_firmware);

	return ret;
}

static gboolean
ip_retry (gpointer user_data)
{
	NMDevice *device = NM_DEVICE (user_data);
	NMDeviceBondPrivate *priv = NM_DEVICE_BOND_GET_PRIVATE (device);
	const char *iface = nm_device_get_ip_iface (device);

	priv->ip_retry_id = 0;

	if (nm_device_get_state (device) != NM_DEVICE_STATE_IP_CONFIG)
		return FALSE;

	if (priv->carrier_waiting == FALSE) {
		if (priv->ip4_waiting) {
			nm_log_info (LOGD_DEVICE | LOGD_IP4, "(%s): retrying IPv4 config", iface);
			priv->ip4_waiting = FALSE;
			nm_device_activate_stage3_ip4_start (device);
		}

		if (priv->ip6_waiting) {
			nm_log_info (LOGD_DEVICE | LOGD_IP6, "(%s): retrying IPv6 config", iface);
			priv->ip6_waiting = FALSE;
			nm_device_activate_stage3_ip6_start (device);
		}
	}

	return FALSE;
}

static void
ip_schedule_retry (NMDevice *device)
{
	NMDeviceBondPrivate *priv = NM_DEVICE_BOND_GET_PRIVATE (device);

	if (priv->ip_retry_id == 0)
		priv->ip_retry_id = g_idle_add (ip_retry, device);
}

static gboolean
enslave_slave (NMDevice *device, NMDevice *slave, NMConnection *connection)
{
	const char *iface = nm_device_get_ip_iface (device);
	const char *slave_iface = nm_device_get_ip_iface (slave);
	gboolean success, no_firmware = FALSE;

	nm_device_hw_take_down (slave, TRUE);

	success = nm_system_bond_enslave (nm_device_get_ip_ifindex (device),
	                                  iface,
	                                  nm_device_get_ip_ifindex (slave),
	                                  slave_iface);

	nm_device_hw_bring_up (slave, TRUE, &no_firmware);

	if (success) {
		nm_log_info (LOGD_DEVICE, "(%s): enslaved bond slave %s", iface, slave_iface);
		g_object_notify (G_OBJECT (device), "slaves");

		/* If waiting for a slave to continue with IP config, start now */
		ip_schedule_retry (device);
	}

	return success;
}

static gboolean
release_slave (NMDevice *device, NMDevice *slave)
{
	gboolean success, no_firmware = FALSE;

	success = nm_system_bond_release (nm_device_get_ip_ifindex (device),
	                                  nm_device_get_ip_iface (device),
	                                  nm_device_get_ip_ifindex (slave),
	                                  nm_device_get_ip_iface (slave));
	nm_log_info (LOGD_DEVICE, "(%s): released bond slave %s (success %d)",
	             nm_device_get_ip_iface (device),
	             nm_device_get_ip_iface (slave),
	             success);
	g_object_notify (G_OBJECT (device), "slaves");

	/* Kernel bonding code "closes" the slave when releasing it, (which clears
	 * IFF_UP), so we must bring it back up here to ensure carrier changes and
	 * other state is noticed by the now-released slave.
	 */
	if (!nm_device_hw_bring_up (slave, TRUE, &no_firmware)) {
		nm_log_warn (LOGD_DEVICE, "(%s): released bond slave could not be brought up.",
		             nm_device_get_iface (slave));
	}

	return success;
}

static void
carrier_changed (NMDevice *device,
                 GParamSpec *pspec,
                 gpointer user_data)
{
	NMDeviceBondPrivate *priv = NM_DEVICE_BOND_GET_PRIVATE (device);

	if (priv->carrier_waiting && nm_device_wired_get_carrier (NM_DEVICE_WIRED (device))) {
		priv->carrier_waiting = FALSE;
		ip_schedule_retry (device);
	}
}

static gboolean
have_any_ready_slaves (NMDevice *device)
{
	GSList *slaves, *iter;
	gboolean usable = FALSE;

	slaves = nm_device_master_get_slaves (device);
	for (iter = slaves; iter; iter = g_slist_next (iter)) {
		if (nm_device_get_enslaved (iter->data)) {
			usable = TRUE;
			break;
		}
	}
	g_slist_free (slaves);

	return usable;
}

static gboolean
ip4_requires_slaves (NMDevice *device)
{
	NMConnection *connection;
	NMSettingIP4Config *s_ip4;
	const char *method = NULL;

	connection = nm_device_get_connection (device);
	g_assert (connection);

	s_ip4 = nm_connection_get_setting_ip4_config (connection);
	if (s_ip4)
		method = nm_setting_ip4_config_get_method (s_ip4);

	return g_strcmp0 (method, NM_SETTING_IP4_CONFIG_METHOD_AUTO) == 0;
}

static NMActStageReturn
act_stage3_ip4_config_start (NMDevice *device, NMDeviceStateReason *reason)
{
	NMDeviceBondPrivate *priv = NM_DEVICE_BOND_GET_PRIVATE (device);

	priv->ip4_waiting = FALSE;

	if (ip4_requires_slaves (device)) {
		if (have_any_ready_slaves (device) == FALSE) {
			priv->ip4_waiting = TRUE;
			nm_log_info (LOGD_DEVICE | LOGD_IP4, "(%s): IPv4 config waiting until slaves are present",
			             nm_device_get_ip_iface (device));
		}

		if (!nm_device_wired_get_carrier (NM_DEVICE_WIRED (device))) {
			priv->carrier_waiting = TRUE;
			nm_log_info (LOGD_DEVICE | LOGD_IP4, "(%s): IPv4 config waiting until carrier is on",
			             nm_device_get_ip_iface (device));
		}

		if (priv->ip4_waiting || priv->carrier_waiting)
			return NM_ACT_STAGE_RETURN_POSTPONE;
	}

	/* Proceed with normal IPv4 configuration */
	return NM_DEVICE_CLASS (nm_device_bond_parent_class)->act_stage3_ip4_config_start (device, reason);
}

static gboolean
ip6_requires_slaves (NMDevice *device)
{
	NMConnection *connection;
	NMSettingIP6Config *s_ip6;
	const char *method = NULL;

	connection = nm_device_get_connection (device);
	g_assert (connection);

	s_ip6 = (NMSettingIP6Config *) nm_connection_get_setting (connection, NM_TYPE_SETTING_IP6_CONFIG);
	if (s_ip6)
		method = nm_setting_ip6_config_get_method (s_ip6);

	/* SLAAC, DHCP, and Link-Local depend on connectivity (and thus slaves)
	 * to complete addressing.  SLAAC and DHCP obviously need a peer to
	 * provide a prefix, while Link-Local must perform DAD on the local link.
	 */
	return    g_strcmp0 (method, NM_SETTING_IP6_CONFIG_METHOD_AUTO) == 0
	       || g_strcmp0 (method, NM_SETTING_IP6_CONFIG_METHOD_DHCP) == 0
	       || g_strcmp0 (method, NM_SETTING_IP6_CONFIG_METHOD_LINK_LOCAL) == 0;
}

static NMActStageReturn
act_stage3_ip6_config_start (NMDevice *device, NMDeviceStateReason *reason)
{
	NMDeviceBondPrivate *priv = NM_DEVICE_BOND_GET_PRIVATE (device);

	priv->ip6_waiting = FALSE;

	if (ip6_requires_slaves (device)) {
		if (have_any_ready_slaves (device) == FALSE) {
			priv->ip6_waiting = TRUE;
			nm_log_info (LOGD_DEVICE | LOGD_IP6, "(%s): IPv6 config waiting until slaves are present",
			             nm_device_get_ip_iface (device));
		}

		if (!nm_device_wired_get_carrier (NM_DEVICE_WIRED (device))) {
			priv->carrier_waiting = TRUE;
			nm_log_info (LOGD_DEVICE | LOGD_IP6, "(%s): IPv6 config waiting until carrier is on",
			             nm_device_get_ip_iface (device));
		}

		if (priv->ip6_waiting || priv->carrier_waiting)
			return NM_ACT_STAGE_RETURN_POSTPONE;
	}

	/* Proceed with normal IPv6 configuration */
	return NM_DEVICE_CLASS (nm_device_bond_parent_class)->act_stage3_ip6_config_start (device, reason);
}

static void
deactivate (NMDevice *device)
{
	NMDeviceBondPrivate *priv = NM_DEVICE_BOND_GET_PRIVATE (device);

	priv->ip4_waiting = FALSE;
	priv->ip6_waiting = FALSE;
	priv->carrier_waiting = FALSE;

	if (NM_DEVICE_CLASS (nm_device_bond_parent_class)->deactivate)
		NM_DEVICE_CLASS (nm_device_bond_parent_class)->deactivate (device);
}

/******************************************************************/

NMDevice *
nm_device_bond_new (const char *udi, const char *iface)
{
	g_return_val_if_fail (udi != NULL, NULL);
	g_return_val_if_fail (iface != NULL, NULL);

	return (NMDevice *) g_object_new (NM_TYPE_DEVICE_BOND,
	                                  NM_DEVICE_INTERFACE_UDI, udi,
	                                  NM_DEVICE_INTERFACE_IFACE, iface,
	                                  NM_DEVICE_INTERFACE_DRIVER, "bonding",
	                                  NM_DEVICE_INTERFACE_TYPE_DESC, "Bond",
	                                  NM_DEVICE_INTERFACE_DEVICE_TYPE, NM_DEVICE_TYPE_BOND,
	                                  NULL);
}

static void
constructed (GObject *object)
{
	G_OBJECT_CLASS (nm_device_bond_parent_class)->constructed (object);

	nm_log_dbg (LOGD_HW | LOGD_DEVICE, "(%s): kernel ifindex %d",
	            nm_device_get_iface (NM_DEVICE (object)),
	            nm_device_get_ifindex (NM_DEVICE (object)));

	g_signal_connect (object,
	                  "notify::" NM_DEVICE_BOND_CARRIER,
	                  G_CALLBACK (carrier_changed),
	                  NULL);
}

static void
nm_device_bond_init (NMDeviceBond * self)
{
	g_signal_connect (self, "state-changed", G_CALLBACK (device_state_changed), NULL);
}

static void
get_property (GObject *object, guint prop_id,
              GValue *value, GParamSpec *pspec)
{
	const guint8 *current_addr;
	GPtrArray *slaves;
	GSList *list, *iter;

	switch (prop_id) {
	case PROP_HW_ADDRESS:
		current_addr = nm_device_wired_get_hwaddr (NM_DEVICE_WIRED (object));
		g_value_take_string (value, nm_utils_hwaddr_ntoa (current_addr, ARPHRD_ETHER));
		break;
	case PROP_CARRIER:
		g_value_set_boolean (value, nm_device_wired_get_carrier (NM_DEVICE_WIRED (object)));
		break;
	case PROP_SLAVES:
		slaves = g_ptr_array_new ();
		list = nm_device_master_get_slaves (NM_DEVICE (object));
		for (iter = list; iter; iter = iter->next)
			g_ptr_array_add (slaves, g_strdup (nm_device_get_path (NM_DEVICE (iter->data))));
		g_slist_free (list);
		g_value_take_boxed (value, slaves);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
set_property (GObject *object, guint prop_id,
			  const GValue *value, GParamSpec *pspec)
{
	switch (prop_id) {
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nm_device_bond_class_init (NMDeviceBondClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	NMDeviceClass *parent_class = NM_DEVICE_CLASS (klass);
	NMDeviceWiredClass *wired_class = NM_DEVICE_WIRED_CLASS (klass);

	g_type_class_add_private (object_class, sizeof (NMDeviceBondPrivate));

	/* virtual methods */
	object_class->constructed = constructed;
	object_class->get_property = get_property;
	object_class->set_property = set_property;

	parent_class->get_generic_capabilities = real_get_generic_capabilities;
	parent_class->update_hw_address = real_update_hw_address;
	parent_class->is_available = is_available;
	parent_class->get_best_auto_connection = real_get_best_auto_connection;
	parent_class->check_connection_compatible = real_check_connection_compatible;

	parent_class->spec_match_list = spec_match_list;
	parent_class->connection_match_config = connection_match_config;

	parent_class->act_stage1_prepare = real_act_stage1_prepare;
	parent_class->act_stage3_ip4_config_start = act_stage3_ip4_config_start;
	parent_class->act_stage3_ip6_config_start = act_stage3_ip6_config_start;
	parent_class->deactivate = deactivate;
	parent_class->enslave_slave = enslave_slave;
	parent_class->release_slave = release_slave;

	wired_class->carrier_action = carrier_action;

	/* properties */
	g_object_class_install_property
		(object_class, PROP_HW_ADDRESS,
		 g_param_spec_string (NM_DEVICE_BOND_HW_ADDRESS,
							  "Active MAC Address",
							  "Currently set hardware MAC address",
							  NULL,
							  G_PARAM_READABLE));

	g_object_class_install_property
		(object_class, PROP_CARRIER,
		 g_param_spec_boolean (NM_DEVICE_BOND_CARRIER,
							   "Carrier",
							   "Carrier",
							   FALSE,
							   G_PARAM_READABLE));

	g_object_class_install_property
		(object_class, PROP_SLAVES,
		 g_param_spec_boxed (NM_DEVICE_BOND_SLAVES,
		                     "Slaves",
		                     "Slaves",
		                     DBUS_TYPE_G_ARRAY_OF_OBJECT_PATH,
		                     G_PARAM_READABLE));

	/* Signals */
	signals[PROPERTIES_CHANGED] =
		nm_properties_changed_signal_new (object_class,
										  G_STRUCT_OFFSET (NMDeviceBondClass, properties_changed));

	dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (klass),
									 &dbus_glib_nm_device_bond_object_info);

	dbus_g_error_domain_register (NM_BOND_ERROR, NULL, NM_TYPE_BOND_ERROR);
}
