/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/* NetworkManager -- Network link manager
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Copyright (C) 2005 - 2014 Red Hat, Inc.
 * Copyright (C) 2006 - 2008 Novell, Inc.
 * Copyright (C) 2011 Intel Corporation. All rights reserved.
 */

#include <config.h>

#include <errno.h>
#include <string.h>

#include <linux/nl80211.h>
#include <netlink/genl/ctrl.h>
#include <netlink/genl/genl.h>

#include "nm-netlink-monitor.h"
#include "nm-logging.h"

static int
ack_handler (struct nl_msg *msg, void *arg)
{
	int *done = arg;
	*done = 1;
	return NL_STOP;
}

static int
finish_handler (struct nl_msg *msg, void *arg)
{
	int *done = arg;
	*done = 1;
	return NL_SKIP;
}

static int
error_handler (struct sockaddr_nl *nla, struct nlmsgerr *err, void *arg)
{
	int *done = arg;
	*done = err->error;
	return NL_SKIP;
}

static struct nl_msg *
_nl80211_alloc_msg (int id, int ifindex, int phy, guint32 cmd, guint32 flags)
{
	struct nl_msg *msg;

	msg = nlmsg_alloc ();
	if (msg) {
		genlmsg_put (msg, 0, 0, id, 0, flags, cmd, 0);
		NLA_PUT_U32 (msg, NL80211_ATTR_IFINDEX, ifindex);
		if (phy != -1)
			NLA_PUT_U32 (msg, NL80211_ATTR_WIPHY, phy);
	}
	return msg;

 nla_put_failure:
	nlmsg_free (msg);
	return NULL;
}

/* NOTE: this function consumes 'msg' */
static int
_nl80211_send_and_recv (struct nl_handle *nlh,
                        struct nl_cb *nl_cb,
                        struct nl_msg *msg,
                        int (*valid_handler) (struct nl_msg *, void *),
                        void *valid_data)
{
	struct nl_cb *cb;
	int err, done;

	g_return_val_if_fail (msg != NULL, -ENOMEM);

	cb = nl_cb_clone (nl_cb);
	if (!cb) {
		err = -ENOMEM;
		goto out;
	}

	err = nl_send_auto_complete (nlh, msg);
	if (err < 0)
		goto out;

	done = 0;
	nl_cb_err (cb, NL_CB_CUSTOM, error_handler, &done);
	nl_cb_set (cb, NL_CB_FINISH, NL_CB_CUSTOM, finish_handler, &done);
	nl_cb_set (cb, NL_CB_ACK, NL_CB_CUSTOM, ack_handler, &done);
	if (valid_handler)
		nl_cb_set (cb, NL_CB_VALID, NL_CB_CUSTOM, valid_handler, valid_data);

	/* Loop until one of our NL callbacks says we're done; on success
	 * done will be 1, on error it will be < 0.
	 */
	while (!done) {
		err = nl_recvmsgs (nlh, cb);
		if (err) {
			nm_log_warn (LOGD_WIFI, "nl_recvmsgs() error: (%d) %s",
			             err, nl_geterror ());
			break;
		}
	}
	if (err == 0 && done < 0)
		err = done;

 out:
	nl_cb_put (cb);
	nlmsg_free (msg);
	return err;
}

struct nl80211_wowlan_info {
	gboolean enabled;
};

static int
nl80211_wowlan_handler (struct nl_msg *msg, void *arg)
{
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data (nlmsg_hdr (msg));
	gboolean *enabled = arg;

	if (nla_parse (tb,
	               NL80211_ATTR_MAX,
	               genlmsg_attrdata (gnlh, 0),
	               genlmsg_attrlen (gnlh, 0),
	               NULL) < 0)
		return NL_SKIP;

	if (tb[NL80211_ATTR_WOWLAN_TRIGGERS])
		*enabled = TRUE;

	return NL_SKIP;
}

static int
nl80211_wiphy_info_handler (struct nl_msg *msg, void *arg)
{
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data (nlmsg_hdr (msg));
	int *phy = arg;

	if (nla_parse (tb,
	               NL80211_ATTR_MAX,
	               genlmsg_attrdata (gnlh, 0),
	               genlmsg_attrlen (gnlh, 0),
	               NULL) < 0)
		return NL_SKIP;

	if (tb[NL80211_ATTR_WIPHY])
		*phy = nla_get_u32 (tb[NL80211_ATTR_WIPHY]);

	return NL_SKIP;
}

static gboolean
wifi_nl80211_get_wake_on_lan (int ifindex)
{
	struct nl_handle *nlh;
	struct nl_cb *cb;
	struct nl_msg *msg;
	int id, phy;
	gboolean enabled = FALSE;

	cb = nl_cb_alloc (NL_CB_DEFAULT);
	nlh = nl_handle_alloc_cb (cb);
	nl_cb_put (cb);
	if (!nlh) {
		nm_log_warn (LOGD_WIFI, "unable to allocate netlink handle for checking WoWLAN status: %s",
		             nl_geterror ());
		return FALSE;
	}
	if (genl_connect (nlh) < 0) {
		nm_log_warn (LOGD_WIFI, "unable to connect to netlink handle for checking WoWLAN status: %s",
		             nl_geterror ());
		return FALSE;
	}

	id = genl_ctrl_resolve (nlh, "nl80211");
	g_return_val_if_fail (id >= 0, FALSE);

	cb = nl_cb_alloc (NL_CB_DEFAULT);
	g_return_val_if_fail (cb != NULL, FALSE);

	msg = _nl80211_alloc_msg (id, ifindex, -1, NL80211_CMD_GET_WIPHY, 0);
	if (_nl80211_send_and_recv (nlh, cb, msg, nl80211_wiphy_info_handler, &phy) < 0)
		goto done;

	msg = _nl80211_alloc_msg (id, ifindex, phy, NL80211_CMD_GET_WOWLAN, 0);
	_nl80211_send_and_recv (nlh, cb, msg, nl80211_wowlan_handler, &enabled);

 done:
	nl_cb_put (cb);
	return enabled;
}

